﻿using System;
using System.Collections.Generic;
using System.Globalization;
using System.Text;

namespace RegexNumRangeTool
{
    internal static class RegexNumRangeGen
    {
        public static string Generate(string a, string b)
        {
            if (a == null)
            {
                throw new ArgumentNullException("a");
            }

            if (b == null)
            {
                throw new ArgumentNullException("b");
            }

            if (a.Length == 0 || TextHelper.IsDigit(a) != -1)
            {
                throw new ArgumentException("A is not a number.", "a");
            }

            if (b.Length == 0 || TextHelper.IsDigit(b) != -1)
            {
                throw new ArgumentException("B is not a number.", "b");
            }

            a = a.Remove(0, TextHelper.FindRep('0', a, 0, a.Length - 2));
            b = b.Remove(0, TextHelper.FindRep('0', b, 0, b.Length - 2));

            if (TextHelper.Compare(a, b) > -1)
            {
                string c = a;
                a = b;
                b = c;
            }

            return BuildRegex(Divide(new Range() { A = a, B = b }));
        }

        private static List<Range> Divide(Range fullRange)
        {
            int diff = TextHelper.Compare(fullRange.B, fullRange.A);
            List<Range> ranges = new List<Range>();

            if (diff == -1)
            {
                ranges.Add(fullRange);
            }
            else
            {
                List<Range> bigRanges = new List<Range>();

                for (int i = fullRange.A.Length; i <= fullRange.B.Length; i++)
                {
                    bigRanges.Add(new Range() { A = ((i == fullRange.A.Length) ? fullRange.A : "1" + new string('0', i - 1)), B = ((i == fullRange.B.Length) ? fullRange.B : new string('9', i)), IsBig = true });
                }

                Range range = bigRanges[0];

                {
                    int len = range.A.Length - 1;

                    int x = 1 + TextHelper.FindRepRight('0', range.A, len, (diff == 0) ? 1 : diff);
                    int y = range.A.Length;

                    if (diff > 0)
                    {
                        y -= diff;
                    }

                    string a = range.A;

                    for (int i = x; i <= y; i++)
                    {
                        int b = i - 1;

                        if (i > x)
                        {
                            a = String.Concat(new string[] { a.Substring(0, a.Length - b - 1), ((char)(a[len - b] + 1)).ToString(), new string('0', b) });
                        }

                        i += TextHelper.FindRepRight('9', range.A, len - i, 0);

                        if (i > y)
                        {
                            i -= 1;
                        }

                        ranges.Add(new Range() { A = a, B = range.A.Substring(0, range.A.Length - i) + new string('9', i) });
                    }
                }

                {
                    int len = bigRanges.Count - 1;

                    for (int i = 1; i < len; i++)
                    {
                        ranges.Add(bigRanges[i]);
                    }
                }

                range = (diff == 0) ?
                    bigRanges[bigRanges.Count - 1] :
                    (ranges.Count == 0) ?
                    fullRange : new Range() { A = String.Concat(new string[] { fullRange.A.Substring(0, diff - 1), ((char)(fullRange.A[diff - 1] + 1)).ToString(), new string('0', fullRange.A.Length - diff) }), B = fullRange.B };

                if (range.A == range.B)
                {
                    ranges.Add(range);
                }
                else
                {
                    int x = TextHelper.Compare(range.B, range.A);
                    int y = range.B.Length - TextHelper.FindRepRight('9', range.B, range.B.Length - 1, x);
                    string a = range.A;

                    for (int i = x; i <= y; i++)
                    {
                        if (i > x)
                        {
                            a = range.B.Substring(0, i - 1) + new string('0', range.B.Length - i + 1);
                        }

                        i += TextHelper.FindRep('0', range.B, i - 1, y - 1);

                        if (i > y)
                        {
                            i -= 1;
                        }

                        ranges.Add(new Range() { A = a, B = ((i == y) ? range.B : String.Concat(new string[] { range.B.Substring(0, i - 1), ((char)(range.B[i - 1] - 1)).ToString(), new string('9', range.B.Length - i) })) });
                    }
                }
            }

            return ranges;
        }

        private static string BuildRegex(List<Range> ranges)
        {
            StringBuilder sb = new StringBuilder();

            if (ranges.Count > 1)
            {
                sb.Append('(');
            }

            int rangesLen = ranges.Count - 1;

            for (int rangesIndex = 0; rangesIndex <= rangesLen; rangesIndex++)
            {
                if (rangesIndex != 0)
                {
                    sb.Append('|');
                }

                Range range = ranges[rangesIndex];

                int length = range.A.Length;

                for (int index = 0; index < length; index++)
                {
                    char c1 = range.A[index];
                    char c2 = range.B[index];

                    if (c1 == c2)
                    {
                        sb.Append(c1);
                    }
                    else
                    {
                        sb.Append('[');
                        sb.Append(c1);

                        if (c2 - c1 != 1)
                        {
                            sb.Append('-');
                        }

                        sb.Append(c2);
                        sb.Append(']');

                        if (c1 == '0' && c2 == '9')
                        {
                            int oldRangesIndex = rangesIndex;

                            while (true)
                            {
                                if (rangesIndex == rangesLen)
                                {
                                    Range newRange = ranges[rangesIndex];

                                    if (rangesIndex != oldRangesIndex && TextHelper.FindRep('9', newRange.B, 0, newRange.B.Length - 1) != newRange.B.Length)
                                    {
                                        rangesIndex--;
                                    }

                                    break;
                                }
                                else if (!ranges[rangesIndex].IsBig)
                                {
                                    if (rangesIndex != oldRangesIndex)
                                    {
                                        rangesIndex--;
                                    }

                                    break;
                                }

                                rangesIndex++;
                            }

                            int q1 = length - index;
                            int q2 = q1 + (rangesIndex - oldRangesIndex);

                            if (q1 > 1 || q2 > 1)
                            {
                                sb.Append('{');
                                sb.Append(q1.ToString(CultureInfo.InvariantCulture));

                                if (q2 > q1)
                                {
                                    sb.Append(',');
                                    sb.Append(q2.ToString(CultureInfo.InvariantCulture));
                                }

                                sb.Append('}');
                            }

                            break;
                        }
                    }
                }
            }

            if (ranges.Count > 1)
            {
                sb.Append(')');
            }

            return sb.ToString();
        }

        private struct Range
        {
            public string A;
            public string B;
            public bool IsBig;
        }

        private static class TextHelper
        {
            private static int[] charmap = new int[]
            {
                1, 1, 1, 1, 1, 1, 1, 1, 1, 3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 2, 0, 0, 0, 0, 0, 0, 0,
                0, 0, 0, 0, 0, 0, 0, 0, 36, 36, 36, 36, 36, 36, 36, 36, 36, 36, 0, 0,
                0, 0, 0, 0, 0, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40,
                40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 40, 0, 0, 0, 0, 32, 0, 48, 48, 48,
                48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48, 48,
                48, 48, 48, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
                1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1
            };
            private static int charmapLength = 160;

            public static bool IsDigit(char c)
            {
                return ((c < charmapLength) && ((charmap[c] & 4) != 0));
            }

            public static int IsDigit(string s)
            {
                return Validate(s, IsDigit);
            }

            public static int Validate(string str, Func<char, bool> validator)
            {
                int length = str.Length;

                for (int index = 0; index < length; index++)
                {
                    if (!validator.Invoke(str[index]))
                    {
                        return index;
                    }
                }

                return -1;
            }

            public static int Compare(string str1, string str2)
            {
                if (str1.Length > str2.Length)
                {
                    return 0;
                }

                if (str1.Length == str2.Length)
                {
                    int length = str1.Length;

                    for (int index = 0; index < length; index++)
                    {
                        if (str1[index] > str2[index])
                        {
                            return index + 1;
                        }

                        if (str1[index] < str2[index])
                        {
                            return -3;
                        }
                    }

                    return -1;
                }

                return -2;
            }

            public static int FindRep(char chr, string str, int beginPos, int endPos)
            {
                int pos;

                for (pos = beginPos; pos <= endPos; pos++)
                {
                    if (str[pos] != chr)
                    {
                        break;
                    }
                }

                return pos - beginPos;
            }

            public static int FindRepRight(char chr, string str, int beginPos, int endPos)
            {
                int pos;

                for (pos = beginPos; pos >= endPos; pos--)
                {
                    if (str[pos] != chr)
                    {
                        break;
                    }
                }

                return beginPos - pos;
            }
        }
    }
}
